"""Query Mars RSS API module."""

from urllib.parse import urlencode

from requests import get

from ..misc import logger


API_URL = 'https://mars.nasa.gov/rss/api/'

HEADERS = {
    'Accept': 'application/json, text/javascript, */*; q=0.01',
}

log_api, debug_api = logger('Mars API', info_stdout=True)


def _flat_json(json) -> dict:
    """Flatten JSON dict."""
    res = {}
    for key, value in json.items():
        if isinstance(value, dict):
            res.update(_flat_json(value))
        else:
            res[key] = value
    return res


def get_json(page=0, num=100, category='mars2020', order='sol+desc'):
    """Get images list and metadata as JSON.

    Parameters
    ----------
    page: int, optional
        Page to query (default: `0`).
    num: int, optional
        Number of result per page (default: `100`).
        Currently this value is limited by the API at 100.
    category: str, optional
        Category of images to query (default: `mars2020`).
    order: str, optional
        Images sorting order (default: `sol+desc`).

    Returns
    -------
    [dict, ...]
        List of images infos (flattened).
    int
        Total number of images for the query.

    Note
    ----
    Select only the full-res and RAW images.
    Discard the thumbnails and color-processed images.

    Order by descreasing `sol`.

    """
    params = {
        'feed': 'raw_images',
        'feedtype': 'json',
        'category': category,
        'num': num,
        'page': page,
        'order': order,
        'extended': 'sample_type::full,product_type::raw',
    }

    payload = urlencode(params, safe=':+,')

    log_api.debug('Payload: %s', payload)

    resp = get(API_URL, params=payload, headers=HEADERS).json()

    imgs = [_flat_json(img) for img in resp['images']]
    total = resp['total_results']

    first_img = 1 + page * num
    last_img = min(num, len(imgs)) + page * num
    current_page = page + 1
    last_page = total // num + 1

    log_api.info('Downloaded images %i-%i/%i (page %i/%i)',
                 first_img, last_img, total, current_page, last_page)

    log_api.debug('Response: %s', resp)

    return imgs, total


def get_all(lastest=None, **kwargs):
    """Get the list of all images since the last known image.

    Parameters
    ----------
    lastest: str, optional
        Latest image id.
        If None is provided (default), the full list
        images will be downloaded.

    Returns
    -------
    [dict, ...]
        List of the latest images.

    See Also
    --------
    :py:func:`get_json`

    """
    new_imgs, total = get_json(page=0, **kwargs)

    last_page = total // len(new_imgs) + 1

    imgs = []
    for page in range(1, last_page + 1):
        imgs_id = [img['imageid'] for img in new_imgs]

        if lastest is not None and lastest in imgs_id:
            imgs.extend(new_imgs[:imgs_id.index(lastest)])
            break

        imgs.extend(new_imgs)

        if page < last_page:
            new_imgs, _ = get_json(page=page, **kwargs)

    return imgs[::-1]



"""Mars Rover database module."""

from pathlib import Path
from sqlite3 import connect

from IPython.core.display import HTML

from ..api import get_all
from ..image import RGB
from ..misc import logger, now


log_db, debug_db = logger('Database')


class RoverDatabase:
    """Rover database object.

    Parameters
    ----------
    fname: str or pathlib.Path
        Database filename.
    category: str
        API category name.

    """
    FIELDS = [
        ('instrument', 'TEXT'),
        ('sol', 'INTEGER'),
        ('imageid', 'TEXT'),
        ('date_taken_utc', 'TEXT'),
        ('date_taken_mars', 'TEXT'),
        ('date_received', 'TEXT'),
        ('sclk', 'TEXT'),
        ('dimension', 'TEXT'),
        ('filter_name', 'TEXT'),
        ('xyz', 'TEXT'),
        ('attitude', 'TEXT'),
        ('camera_vector', 'TEXT'),
        ('camera_model_component_list', 'TEXT'),
        ('camera_position', 'TEXT'),
        ('camera_model_type', 'TEXT'),
        ('subframeRect', 'TEXT'),
        ('scaleFactor', 'TEXT'),
        ('mastAz', 'TEXT'),
        ('mastEl', 'TEXT'),
        ('title', 'TEXT'),
        ('caption', 'TEXT'),
        ('sample_type', 'TEXT'),
        ('json_link', 'TEXT'),
        ('link', 'TEXT'),
        ('small', 'TEXT'),
        ('medium', 'TEXT'),
        ('large', 'TEXT'),
        ('full_res', 'TEXT'),
        ('drive', 'TEXT'),
        ('site', 'INTEGER'),
        ('credit', 'TEXT'),
    ]

    def __init__(self, fname, category):
        self.fname = Path(fname)
        self.category = category

        self.conn = connect(self.fname)
        self.cursor = self.conn.cursor()
        self.create_imgs_table()

        log_db.debug('Connect: %s', self.fname)

    def __repr__(self):
        n = self.count
        if n == 0:
            desc = 'EMPTY'
        else:
            desc = f'{n} image' + ('' if n > 1 else '')
            desc += f' | latest: {self.latest_img}'

        return f'<{self.__class__.__name__}> {desc}'

    def __call__(self, *args, show=False, **kwargs):
        if show:
            return self.show(**kwargs)

        return self.search(*args, **kwargs)

    def __contains__(self, other):
        return self.contains(other)

    def create_imgs_table(self):
        """Create Imgs table."""
        log_db.debug('Create tables (if not exists)')

        # Create Img table
        fields = ','.join([f'{field} {_type}' for field, _type, in self.FIELDS])
        sql = f'CREATE TABLE IF NOT EXISTS Imgs ({fields}, PRIMARY KEY("imageid"))'
        self.cursor.execute(sql)

        # Create Updates table
        sql = ('CREATE TABLE IF NOT EXISTS Updates ('
               'update_date, '
               'nb_new_imgs, '
               'nb_imgs_before, '
               'nb_imgs_after, '
               'old_latest_img, '
               'new_latest_img)')
        self.cursor.execute(sql)

    def delete_imgs_table(self):
        """Delete Imgs table."""
        log_db.warning('Remove all tables from %s', self.fname)

        sql = 'DROP TABLE IF EXISTS Imgs'
        self.cursor.execute(sql)

    def purge(self):
        """Purge Imgs database."""
        self.delete_imgs_table()
        self.create_imgs_table()

    def row(self, json):
        """Extract data from JSON."""
        default = {key: None for key, _ in self.FIELDS}
        default.update(json)
        return tuple(default.values())

    @property
    def insert_into(self):
        """SQL insert into statement."""
        return 'INSERT INTO Imgs VALUES (' + ','.join(len(self.FIELDS) * '?') + ')'

    def insert(self, json):
        """Insert data values."""
        nb_imgs_before = self.count
        old_latest_img = self.latest_img
        update_date = now()

        log_db.debug('Before: %i images | Latest: %s', nb_imgs_before, old_latest_img)

        if isinstance(json, list):
            self.cursor.executemany(self.insert_into, [self.row(img) for img in json])
        else:
            self.cursor.execute(self.insert_into, self.row(json))

        # Commit changes
        self.conn.commit()

        nb_imgs_after = self.count
        new_latest_img = self.latest_img
        nb_new_imgs = nb_imgs_after - nb_imgs_before

        log_db.info('%i images were added.', nb_new_imgs)
        log_db.debug('After: %i images | Latest: %s', nb_imgs_after, new_latest_img)

        # Save changes in update table
        row = tuple([
            update_date,
            nb_new_imgs,
            nb_imgs_before,
            nb_imgs_after,
            old_latest_img,
            new_latest_img,
        ])
        sql = 'INSERT INTO Updates VALUES (' + ','.join(len(row) * '?') + ')'

        self.cursor.execute(sql, row)
        self.conn.commit()

        log_db.debug('Update infos saved')

    def update(self, num=100):
        """Update the database content."""
        json = get_all(lastest=self.latest_img,
                       num=num,
                       category=self.category,
                       order='sol+desc')
        if json:
            self.insert(json)

    @property
    def count(self):
        """Count the number of images in the database."""
        self.cursor.execute('SELECT COUNT(imageid) from Imgs')
        return self.cursor.fetchone()[0]

    @property
    def imgs(self):
        """List of images id in the database."""
        self.cursor.execute('SELECT imageid from Imgs')
        return [img_id for (img_id,) in self.cursor.fetchall()]

    @property
    def latest_img(self):
        """Lastest image id added to the database."""
        self.cursor.execute('SELECT imageid from Imgs ORDER BY rowid DESC LIMIT 1')
        res = self.cursor.fetchone()
        return res[0] if res else None

    def contains(self, img_id):
        """check if a image is in the database."""
        self.cursor.execute(
            'SELECT COUNT(*) from Imgs WHERE imageid=? LIMIT 1', (img_id,))
        return self.cursor.fetchone()[0] == 1

    def search(self, fields='imageid', where=None, order=None, limit=None):
        """Search into the image database."""
        if isinstance(fields, (tuple, list)):
            fields = ', '.join(fields)

        sql = f'SELECT {fields} FROM Imgs'

        if where is not None:
            sql += f' WHERE {where}'

        if order is not None:
            sql += f' ORDER BY {order}'

        if limit is not None:
            sql += f' LIMIT {limit}'

        self.cursor.execute(sql)
        res = self.cursor.fetchall()

        if not res:
            return []

        if len(res[0]) == 1:
            return [data for (data,) in res]

        return res

    def show(self, **kwargs):
        """Show images for a search requests."""
        style = ';'.join([
            'display: flex',
            'flex-direction: row',
            'flex-wrap: wrap',
            'justify-content: space-evenly',
            'align-items: baseline',
        ])
        html = f'<div style="{style}">'

        imgs = self.search(fields='imageid,small,full_res', **kwargs)

        for img_id, thumbnail, url in imgs:
            html += f'<a href="{url}" target="_blank">'
            html += f'<p style="text-align:center;font-size:10px">{img_id}</p>'
            html += f'<img src="{thumbnail}" style="margin:auto"/>'
            html += '</a>'

        html += '</div>'

        return HTML(html)

    def group_rgb(self):
        """Group RGB pairs."""
        sql = '''
SELECT
    rgb_name,
    group_concat(imageid) as image_ids,
    group_concat(channel) as channels,
    group_concat(full_res) as urls,
    count(imageid) AS nb,
    instrument,
    dimension,
    camera_vector,
    sol,
    attitude
FROM (
    SELECT
        *,
        substr(imageid, 3, 1) as channel,
        substr(imageid, 1, 1) || "_F" ||
        substr(imageid, 4) || "-" || substr(imageid, 2, 1) as rgb_name
    FROM Imgs
    WHERE
        camera_vector != "UNK"
    AND
        (channel  = "R" OR channel  = "G" OR channel  = "B")
    ORDER BY
        channel DESC,
        rgb_name
)
GROUP BY camera_vector, sol, attitude
HAVING nb > 1
ORDER BY
    nb DESC,
    sol
'''
        self.cursor.execute(sql)
        return [RGB(res) for res in self.cursor.fetchall()]

    def save_rgb(self, output_dir='RGB'):
        """Save all RGB in `output_dir`."""
        return [rgb.save(output_dir=output_dir) for rgb in self.group_rgb()]



"""Logger module."""

import sys
import logging


class InfoFilter(logging.Filter):
    """Filter only record with level INFO."""
    def filter(self, record):
        return record.levelno == logging.INFO


class NotInfoFilter(logging.Filter):
    """Filter only record not with level INFO."""
    def filter(self, record):
        return record.levelno != logging.INFO


def logger(name, info_stdout=False):
    """Create a custom logger with its debugger."""
    snake_name = name.replace(' ', '_').lower()
    logger = logging.getLogger(f'{__package__}.{snake_name}')

    if logger.hasHandlers():
        logger.handlers.clear()

    stderr = logging.StreamHandler()
    logger.addHandler(stderr)

    fmt = logging.Formatter(f'[{name}] %(message)s')
    stderr.setFormatter(fmt)

    if info_stdout:
        stdout = logging.StreamHandler(sys.stdout)
        stdout.setLevel(logging.INFO)
        stdout.setFormatter(fmt)
        logger.addHandler(stdout)

        # Filter INFO to send them to stdout
        stdout.addFilter(InfoFilter())
        stderr.addFilter(NotInfoFilter())
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    return logger, debug(logger, default=logger.level)


def debug(logger, default=logging.WARNING):
    """Custom debug toggler."""
    def _debug(enabled=True):
        """Change logger level."""
        if enabled == 'info':
            logger.setLevel(logging.INFO)
        elif enabled == 'warning':
            logger.setLevel(logging.WARNING)
        elif enabled:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(default)

        logger.info('Change logger level to %s', logging.getLevelName(logger.level))

    return _debug



"""Mars Rovers Catalogue setup."""

import re

from setuptools import setup, find_packages


with open('catalog/version.py') as f:
    __version__ = re.findall(
        r'__version__ = \'(\d+\.\d+\.\d+)\'',
        f.read()
    )[0]

with open('README.md', encoding='utf-8') as f:
    readme = f.read()


setup(
    name='mars-rovers-catalog',
    version=__version__,
    description='Mars Rovers Catalog toolbox',
    author='Benoit Seignovert',
    author_email='benoit.seignovert@univ-nantes.fr',
    url='https://gitlab.univ-nantes.fr/mars-rovers/catalog/',
    license='BSD',
    python_requires='>=3.6',
    install_requires=[
        'requests',
        'Pillow>=8.0',
        'ipython',
    ],
    packages=find_packages(),
    include_package_data=True,
    keywords=['nasa', 'catalog', 'mars2020', 'perseverance', 'msl', 'curiosity'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Astronomy',
    ],
    long_description=readme,
    long_description_content_type='text/markdown',
)



"""Mars Rover RGB image module."""

from pathlib import Path
from collections import UserDict

from PIL import Image

from requests import get

from ..misc import logger


log_rgb, debug_rgb = logger('RGB', info_stdout=True)


class RGB(UserDict):
    """RGB image object."""

    def __init__(self, args):
        super().__init__()

        name, imgs, channels, urls, nb, instr, dim, camera_vector, sol, attitude = args

        self.name = name
        self.instrument = instr.replace('_LEFT', '').replace('_RIGHT', '')

        self.nb_imgs = nb
        self.imgs = imgs.split(',')
        self.channels = channels.split(',')
        self.urls = urls.split(',')

        self.dim = tuple(map(int, dim[1:-1].split(',')))
        self.camera_vector = camera_vector
        self.sol = sol
        self.attitude = attitude

        for channel, url in zip(self.channels, self.urls):
            self[channel] = url

    def __str__(self):
        return self.name

    def __repr__(self):
        return (f'<{self.__class__.__name__}> '
                f'{self} | {self.instrument} | '
                f'{self.nb_imgs} images | Channels: {",".join(self.channels)}')

    def __getitem__(self, key):
        if key not in 'RGB':
            return self.data[key]

        if key in self:
            return Image.open(get(self.data[key], stream=True).raw).split()[0]

        # Empty blank image if missing
        return Image.new('L', self.dim)

    @property
    def is_complete(self):
        """Check if the RGB image is complete."""
        return 'R' in self and 'G' in self and 'B' in self

    def save(self, output_dir='RGB'):
        """Save RGB image.

        The images will be save as:

            `OUTPUT_DIR/INSTRUMENT/IMG.jpg`

        """
        if not self.is_complete:
            missing = ', '.join({'R', 'G', 'B'} - set(self.keys()))
            log_rgb.warning('Missing channel: `%s` in %s', missing, self.name)
            return None

        fname = (Path(output_dir) / self.instrument / self.name).with_suffix('.jpg')

        if fname.exists():
            log_rgb.debug('%s exists', fname.name)
            return fname

        fname.parent.mkdir(parents=True, exist_ok=True)

        rgb = Image.merge('RGB', (self['R'], self['G'], self['B']))
        rgb.save(fname)

        log_rgb.info('Created: %s', fname.name)
        return fname
